/*
 * Copyright (c) 2009, 2012, 2013, 2014 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Authors: Sugih Jamin, Manoj Rajagopalan
 *
 * Using FBO as accumulation buffer.
 *
*/
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
using namespace std;

#ifdef __APPLE__
#include <GLUT/glut.h>
#define GL_RGBA32F GL_RGBA32F_ARB
#else
#include <GL/glew.h>
#include <GL/glut.h>
#endif

GLdouble width=640.0;
GLdouble height=640.0;   /* window width and height */
int wd;                  /* GLUT window handle */
GLuint fbod;             /* framebuffer object */
GLenum gl_error;

void
kbd(unsigned char key, int x, int y)
{
  switch((char)key) {
  case 'q':
  case 27:    /* ESC */
    glutDestroyWindow(wd);
    exit(0);
  default:
    break;
  }

  return;
}

void
reshape(int w, int h)
{
  width = (GLdouble) w;
  height = (GLdouble) h;

  glViewport(0, 0, (GLsizei) w, (GLsizei) h);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, width, 0.0, height, -1.f, 1.f);

  return;
}

void
disp(void)
{
  GLfloat triad[3][2] = { // (0,0) is the lower left corner
    { 256.0, 384.0 }, 
    { 256.0, 256.0 }, 
    { 384.0, 384.0 }
  }; 
  GLfloat triad2[3][2] = { // (0,0) is the lower left corner
    { 384.0, 384.0 },
    { 256.0, 256.0 }, 
    { 384.0, 256.0 }
  }; 
#define NRBUF 2  // number of render buffers
  GLuint rbod[NRBUF];
  GLint maxsize;

  glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE, &maxsize);
  if (width > maxsize || height > maxsize) {
    cerr << "Window size > max renderbuffer size" << endl;
    exit(-1);
  }
    
  glBindFramebuffer(GL_FRAMEBUFFER, fbod);

  glGenRenderbuffers(NRBUF, rbod);
  // use first rbo as accumulation buffer
  glBindRenderbuffer(GL_RENDERBUFFER, rbod[0]);
  glRenderbufferStorage(GL_RENDERBUFFER, GL_RGBA32F, width, height);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "RenderbufferStorage: [" 
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(-1);
  }
  glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                            GL_RENDERBUFFER, rbod[0]);

  // use second rbo as depth and stencil buffers
  glBindRenderbuffer(GL_RENDERBUFFER, rbod[1]);
  glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_STENCIL, width, height);
  glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT,
                            GL_RENDERBUFFER, rbod[1]);
  if ((gl_error = glCheckFramebufferStatus(GL_FRAMEBUFFER)) !=
      GL_FRAMEBUFFER_COMPLETE) {
    cerr << "Framebuffer incomplete [0x" << hex << gl_error << "] "
         << glGetString(gl_error) << endl;
    exit(-1);
  }

  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
 
#define NREND 8 // how many samples per pixel
  float weight = 1.0/(float)NREND;

  glEnable(GL_BLEND);
  glBlendColor(0.0, 0.0, 0.0, weight);
  glBlendFunc(GL_CONSTANT_ALPHA, GL_ONE);
  glBlendEquation(GL_FUNC_ADD);

  for (int i = 0; i < NREND; i++) {

    // the first two samples of the first triangle have lighter color 
    if (i == 0) {
      glColor3f(0.12, 0.24, 0.21);
    } else if (i == 1) {
      glColor3f(0.12, 0.24, 0.21);
    } else {
      glColor3f(0.4, 0.8, 0.7);
    }
    glVertexPointer(2, GL_FLOAT, 0, triad); 
    glDrawArrays(GL_TRIANGLES, 0, 3);

    // all samples of the second triangle are of the same color
    glColor3f(0.4, 0.8, 0.7);
    glVertexPointer(2, GL_FLOAT, 0, triad2);
    glDrawArrays(GL_TRIANGLES, 0, 3);

    // so second triangle should be brighter than the first one
  }

  glDisable(GL_BLEND);

  glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
  glBindFramebuffer(GL_READ_FRAMEBUFFER, fbod);

  // now blit the "accumulation buffer" to default framebuffer
  glBlitFramebuffer(0, 0, width, height, 0, 0, width, height,
                    GL_COLOR_BUFFER_BIT, GL_NEAREST);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "BlitFramebuffer: [" 
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(-1);
  }
  glBindFramebuffer(GL_FRAMEBUFFER, 0);

  glDeleteRenderbuffers(NRBUF, rbod);

  glFlush();

  return;
}

void
init_shapes(void)
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glEnableClientState(GL_VERTEX_ARRAY); // allow OpenGL server access to client memory
  //glEnableClientState(GL_COLOR_ARRAY);  // allow OpenGL server access to client memory
  glGenFramebuffers(1, &fbod);

  return;
}

int
main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_SINGLE | GLUT_RGBA);
  glutInitWindowSize((int) width, (int) height);
  wd = glutCreateWindow("Experiment with line drawing" /* title */ );
  
#ifndef __APPLE__
  GLenum err = glewInit();
  if (GLEW_OK != err) {
    /* Problem: glewInit failed, something is seriously wrong. */
    cerr << "Error: " << glewGetErrorString(err) << endl;
  }
  cout <<  "Status: Using GLEW " << glewGetString(GLEW_VERSION) << endl;
#endif /* __APPLE__ */

  glutDisplayFunc(disp);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(kbd);
  
  init_shapes();

  glutMainLoop();
  
  exit(0);
}
