/*
 * Copyright (c) 2009, 2012, 2013, 2014 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Authors: Sugih Jamin, Manoj Rajagopalan
 * VAO without VBO (see vao-vbo.cpp for example of VAO using VBO)
 * and without shaders.  Supported by Mac OS X but not any other
*  graphics card drivers.
 *
*/
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
using namespace std;

#ifdef __APPLE__
#include <GLUT/glut.h>
#define glGenVertexArrays glGenVertexArraysAPPLE
#define glBindVertexArray glBindVertexArrayAPPLE
#else
#include <GL/glew.h>  // must come before the other GL includes
#include <GL/glut.h>
#endif

static GLenum gl_error;

GLdouble width=1280.0;
GLdouble height=640.0;   /* window width and height */
int wd;                  /* GLUT window handle */

enum { TRIAD=0, QUAD, NumVAOs };

GLuint vaods[NumVAOs];        // variable to hold vao descriptor(s)

GLfloat triad[3][2] = { // (0,0) is the lower left corner
  { 0.0, 64.0 }, 
  { 0.0, 0.0 }, 
  { 64.0, 0.0 } 
}; 
GLfloat quad[4][2] = { // (0,0) is the lower left corner
  { 0.0, 64.0 },  // v0
  { 0.0, 0.0 },   // v1
  { 64.0, 64.0 }, // v2; T0: v0-v1-v2
  { 64.0, 0.0 }   // v3; T1: v2-v1-v3
}; 
GLfloat colors[4][3] = {
  { 1.0, 0.0, 0.0 }, 
  { 0.0, 1.0, 0.0 }, 
  { 0.0, 0.0, 1.0 },
  { 0.0, 1.0, 1.0 }
}; 

void
kbd(unsigned char key, int x, int y)
{
  switch((char)key) {
  case 'q':
  case 27:    /* ESC */
    glutDestroyWindow(wd);
    exit(0);
  default:
    break;
  }

  return;
}

void
reshape(int w, int h)
{
  width = (GLdouble) w;
  height = (GLdouble) h;

  glViewport(0, 0, (GLsizei) w, (GLsizei) h);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, width, 0.0, height, -1.f, 1.f);

  return;
}

void
display(void)
{
  int x, y;

  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  for (y=0; y < 10; y++) {
    glLoadIdentity();
    glTranslatef(0.0, y*64.0, 0.0);
    for (x=0; x < 10; x++) {
      glBindVertexArray(vaods[TRIAD]);
      glDrawArrays(GL_TRIANGLES, 0, 3);

      glTranslatef(64.0, 0.0, 0.0);
      glBindVertexArray(vaods[QUAD]);
      glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

      glTranslatef(64.0, 0.0, 0.0);
    }
  }

  glFlush();

  return;
}

void
init_shapes(void)
{
  // Set up vertex-array object (vao) descriptors:
  // grab a number of descriptors, for each descriptor, create and bind a vao, 
  // allow OpenGL access to client memory, and set up vertex array pointer.
  /*
     VAO enables switching between multiple sets of vertex arrays
     (analogous to Texture Unit).
     See Vertex-Array Object section in Redbook 7th ed.
     http://www.informit.com/articles/article.aspx?p=1377833&seqNum=8
  */
  glGenVertexArrays(NumVAOs, vaods); // generate an array of n vao descriptors
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "GenVertArr: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  glBindVertexArray(vaods[TRIAD]); // create and bind vao(s) to use for subsequent
                                 // gl*VertexAttrib* and gl*Draw* calls
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "BindVertArr triad: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  glEnableClientState(GL_VERTEX_ARRAY); // allow OpenGL server access to client memory
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "EnableCS triad VA: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  glVertexPointer(2, GL_FLOAT, 0, (GLvoid *) triad);   // format of data stored in vertex array
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "VertPtr triad: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  glEnableClientState(GL_COLOR_ARRAY);  // allow OpenGL server access to client memory
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "EnableCS triad CA: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  glColorPointer(3, GL_FLOAT, 0, (GLvoid *) colors);    // format of data stored in color array
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "ClrPtr triad: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  glBindVertexArray(vaods[QUAD]);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "BindVertArr quad: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  glEnableClientState(GL_VERTEX_ARRAY);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "EnableCS quad VA: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  glVertexPointer(2, GL_FLOAT, 0, (GLvoid *) quad);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "VertPtr quad: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }


  glEnableClientState(GL_COLOR_ARRAY);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "EnableCS quad CA: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  glColorPointer(3, GL_FLOAT, 0, (GLvoid *) colors);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "ClrPtr quad: [Error "
         << gl_error << "] " << gluGetString(gl_error) << endl;
    exit(1);
  }

  return;
}

int
main(int argc, char *argv[])
{
#ifndef __APPLE__
  cerr << "vao-va: Only supported on Mac OS X" << endl;
  exit(1); 
#endif /* __APPLE__ */

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_SINGLE | GLUT_RGBA);
  glutInitWindowSize((int) width, (int) height);
  wd = glutCreateWindow("Experiment with line drawing" /* title */ );
  
#ifndef __APPLE__
  GLenum err = glewInit();
  if (GLEW_OK != err) {
    /* Problem: glewInit failed, something is seriously wrong. */
    cerr << "Error: " << glewGetErrorString(err) << endl;
  }
  cout <<  "Status: Using GLEW " << glewGetString(GLEW_VERSION) << endl;
#endif /* __APPLE__ */

  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(kbd);
  
  init_shapes();
  
  glutMainLoop();
  
  exit(0);
}
