/*
 * Copyright (c) 2009, 2012, 2013, 2014 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Authors: Sugih Jamin, Manoj Rajagopalan
 * VAO using VBO (see vao-va.cpp for example of VAO without VBO),
 * without shaders.
 *
*/
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
using namespace std;

#ifdef __APPLE__
#include <GLUT/glut.h>
#define glGenVertexArrays glGenVertexArraysAPPLE
#define glBindVertexArray glBindVertexArrayAPPLE
#else
#include "GL/glew.h"  // must come before the other GL includes
#include <GL/glut.h>
#endif

GLdouble width=1280.0;
GLdouble height=640.0;   /* window width and height */
int wd;                   /* GLUT window handle */

enum { TRIAD=0, QUAD, NumVAOs};

GLuint vaods[NumVAOs];        // variable to hold vao descriptor(s)

void
kbd(unsigned char key, int x, int y)
{
  switch((char)key) {
  case 'q':
  case 27:    /* ESC */
    glutDestroyWindow(wd);
    exit(0);
  default:
    break;
  }

  return;
}

void
reshape(int w, int h)
{
  width = (GLdouble) w;
  height = (GLdouble) h;

  glViewport(0, 0, (GLsizei) w, (GLsizei) h);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, width, 0.0, height, -1.f, 1.f);

  return;
}

void
display(void)
{
  int x, y;

  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  for (y=0; y < 10; y++) {
    glLoadIdentity();
    glTranslatef(0.0, y*64.0, 0.0);
    for (x=0; x < 10; x++) {
      glBindVertexArray(vaods[TRIAD]);
      glDrawArrays(GL_TRIANGLES, 0, 3);

      glTranslatef(64.0, 0.0, 0.0);
      glBindVertexArray(vaods[QUAD]);
      glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

      glTranslatef(64.0, 0.0, 0.0);
    }
  }

  glFlush();

  return;
}

void
init_shapes(void)
{
  GLfloat triad[3][2] = { // (0,0) is the lower left corner
    { 0.0, 64.0 }, 
    { 0.0, 0.0 }, 
    { 64.0, 0.0 } 
  }; 
  GLfloat quad[4][2] = { // (0,0) is the lower left corner
    { 0.0, 64.0 },  // v0
    { 0.0, 0.0 },   // v1
    { 64.0, 64.0 }, // v2; T0: v0-v1-v2
    { 64.0, 0.0 }   // v3; T1: v2-v1-v3
  }; 
  GLfloat colors[4][3] = {
    { 1.0, 0.0, 0.0 }, 
    { 0.0, 1.0, 0.0 }, 
    { 0.0, 0.0, 1.0 },
    { 0.0, 1.0, 1.0 }
  }; 

  /* Set up a server-side vertex buffer object (vbo(s)):
     first generate a descriptor for the vbo(s).  Bind
     the descriptor with a buffer object of specified
     type (GL_ARRAY_BUFFER).  Copy the data 
     over, and specify usage hint OpenGL can use to 
     optimize performance.
  */
  GLuint vbods[2];                 // variable to hold vbo descriptor(s)        
  glGenBuffers(2, vbods);      // generate an array of n (=2 here) vbo descriptor

  glBindBuffer(GL_ARRAY_BUFFER, vbods[QUAD]); 
  glBufferData(GL_ARRAY_BUFFER, sizeof(quad)+sizeof(colors), quad, GL_STATIC_DRAW );
  glBufferSubData(GL_ARRAY_BUFFER, sizeof(quad), sizeof(colors), colors);

  glBindBuffer(GL_ARRAY_BUFFER, vbods[TRIAD]); // create a new buffer object, of type
                              // GL_ARRAY_BUFFER, attach it to vbos and make
                              // it the current buffer object
  glBufferData(GL_ARRAY_BUFFER, 
               sizeof(triad)+sizeof(colors),
               triad, GL_STATIC_DRAW );
                              // allocate sizeof(triad) OpenGL server
                              // memory, copy data in "triad" to the newly
                              // allocated space. GL_STATIC_DRAW is a performance
                              // hint to OpenGL that the data is for drawing,
                              // and it will not be changed.
  glBufferSubData(GL_ARRAY_BUFFER, sizeof(triad), sizeof(colors), colors);

  // Set up vertex-array object (vao) descriptors:
  // grab a number of descriptors, for each descriptor, create and bind a vao, 
  // allow OpenGL access to client memory, and set up vertex array pointer.
  /*
     VAO enables switching between multiple sets of vertex arrays
     (analogous to Texture Unit).
     See Vertex-Array Object section in Redbook 7th ed.
     http://www.informit.com/articles/article.aspx?p=1377833&seqNum=8
  */
  glGenVertexArrays(NumVAOs, vaods); // generate an array of n vao descriptors

  glBindVertexArray(vaods[TRIAD]); // create and bind vao(s) to use for subsequent
                                 // gl*VertexAttrib* and gl*Draw* calls

  glEnableClientState(GL_VERTEX_ARRAY); // allow OpenGL server access to client memory
  glEnableClientState(GL_COLOR_ARRAY);  // allow OpenGL server access to client memory
  glVertexPointer(2, GL_FLOAT, 0, 0);   // format of data stored in vertex array
  glColorPointer(3, GL_FLOAT, 0, (GLvoid *) sizeof(triad));    // format of data stored in color array

  glBindVertexArray(vaods[QUAD]);
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_COLOR_ARRAY);
  glBindBuffer(GL_ARRAY_BUFFER, vbods[QUAD]); 
  glVertexPointer(2, GL_FLOAT, 0, 0);
  glColorPointer(3, GL_FLOAT, 0, (GLvoid *) sizeof(quad));

  return;
}

int
main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_SINGLE | GLUT_RGBA);
  glutInitWindowSize((int) width, (int) height);
  wd = glutCreateWindow("Experiment with line drawing" /* title */ );
  
#ifndef __APPLE__
  GLenum err = glewInit();
  if (GLEW_OK != err) {
    /* Problem: glewInit failed, something is seriously wrong. */
    cerr << "Error: " << glewGetErrorString(err) << endl;
  }
  cout <<  "Status: Using GLEW " << glewGetString(GLEW_VERSION) << endl;
#endif /* __APPLE__ */

  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(kbd);
  
  init_shapes();
  
  glutMainLoop();
  
  exit(0);
}
